/*
 ******************************* Creazione file *******************************
 *-------------------------------------|--------------------------------------*
 * Creatore : Marco Savegnago					Data di creazione : 30/03/2000
 * Codice   :
 *
 * Note:
 *
 *
 ******************************* Revisione file *******************************
 *-------------------------------------|--------------------------------------*
 * Revisore :                                  Data di revisione :
 * Versione : 0
 *
 * Note:
 *
 */


#include <vdw.h>
#include <ntddk.h>
#include "NTPCFOMDriver.h"
#include "NTPCFOMDevice.h"
#include "..\NTPCFOMioctl.h"
#include "NTPCFOMHelper.h"

#include "msglog.h"

#pragma warning(disable:4065) // Allow switch statement with no cases

////////////////////////////////////////////////////////////////////////
// NTPCFOMDevice::~NTPCFOMDevice
//
//	Routine Description:
//		This is the destructor for the NTPCFOMDevice
//
//	Parameters:
//		None
//
//	Return Value:
//		None
//
//	Comments:
//		Disconnect and release resources here.
//
//		Although the use of SAFE_DESTRUCTORS in the class definition cures
//		improper emission of the destructor into the INIT section most of the
//		time, certain rare cases can still cause improper behavior.  To avoid
//		these cases, the destructor	must preceed the INIT section, causing it
//		to be referenced first by the default compiler section.
//

NTPCFOMDevice::~NTPCFOMDevice()
{
	ReleaseResources(*m_RegPath);
	delete m_RegPath;
}

#pragma code_seg("INIT")

////////////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::NTPCFOMDevice
//
//	Routine Description:
//		The device constructor is typically responsible for allocating
//		any physical resources that are associated with the device.
//
//	Parameters:
//		Unit - Unit number. This is a number to append to the device's
//			base device name to distinguish multiple units of this
//			device type.
//
//	Return Value:
//		None
//
//	Comments:
//		The device constructor often reads the registry to setup
//		various configurable parameters.

NTPCFOMDevice::NTPCFOMDevice(ULONG Unit) 
:
	KDevice()
{
	if ( ! NT_SUCCESS(m_ConstructorStatus) )
	{
#if DBG
		t << "Failed to create device NTPCFOMDevice unit number " << Unit << " status " << (ULONG) m_ConstructorStatus << EOL;
#endif
		return;
	}

	m_Unit = Unit;

	m_RegPath = CreateRegistryPath(L"NTPCFOMDevice", Unit);
	
	if (m_RegPath == NULL)
	{
		// Error, cannot allocate memory for registry path
#if DBG
		t << "Failed to create registry path\n";
#endif
		m_ConstructorStatus = STATUS_INSUFFICIENT_RESOURCES;
		return;
	}

	// Load data members from the registry
	LoadRegistryParameters();

	// Claim and initialize resources assigned to this device
	m_ConstructorStatus = InitResources();
}

////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::LoadRegistryParams
//
//	Routine Description:
//		Load parameters from the registry and store them in class
//		member variables.
//
//	Parameters:
//		None
//
//	Return Value:
//		None
//
//	Comments:
//		Member variables are updated with values read from registry.
//
//		The registry variables are loaded from the registry path:
//		HKEY_LOCAL_MACHINE\
//		  SYSTEM\
//		    CurrentControlSet\
//			  Services\
//			    NTPCFOMDriver\
//				  Devices\
//				    NTPCFROMDevice\
//				      <unit number>\
//					    <param name>

void NTPCFOMDevice::LoadRegistryParameters(void)
{
	ULONG RegLength;			// Holds length of unsized registry values

	KRegistryKey Params(*m_RegPath);

#if DBG
	t << PWSTR(m_RegPath) << "\n";
#endif

	if (!NT_SUCCESS(Params.LastError()))
	{
		// Error, cannot open device registry key
#if DBG
		t << "!!Failed to create registry key\n";
#endif
		m_ConstructorStatus = STATUS_UNSUCCESSFUL;
		return;
	}

	// Load the registry parameters

	m_IoTimeout = 20000;
	Params.QueryValue(L"IoTimeout", &m_IoTimeout);

#if DBG
	t << "m_IoTimeout loaded from registry, resulting value: [" << m_IoTimeout << "]\n";
#endif

	m_IoEnableIrq = TRUE;
	Params.QueryValue(L"IoEnableIrq", &m_IoEnableIrq);

#if DBG
	t << "m_IoEnableIrq loaded from registry, resulting value: [" << m_IoEnableIrq << "]\n";
#endif

	m_IoBaseAddress = 0x0300;
	Params.QueryValue(L"IoBaseAddress", &m_IoBaseAddress);

#if DBG
	t << "m_IoBaseAddress loaded from registry, resulting value: [" << m_IoBaseAddress << "]\n";
#endif

	m_IoIrq = 0x0000;
	Params.QueryValue(L"IoIrq", &m_IoIrq);

#if DBG
	t << "m_IoIrq loaded from registry, resulting value: [" << m_IoIrq << "]\n";
#endif

	m_FirmwareImage = L"PCCFOM.TSK";
	RegLength = 0;
	Params.QueryValue(L"FirmwareImage", m_FirmwareImage, RegLength, NonPagedPool);
	// Setting RegLength to 0 indicates a new block should be allocated.
	// TODO: If RegLength is returned > 0, a new block was allocated
	// and should be deleted when no longer needed.
#if DBG
	t << "m_FirmwareImage loaded from registry, resulting value: [" << m_FirmwareImage << "]\n";
#endif

	m_InvertCH0_0 = TRUE;
	Params.QueryValue(L"InvertCH0_0", &m_InvertCH0_0);

#if DBG
	t << "m_InvertCH0_0 loaded from registry, resulting value: [" << m_InvertCH0_0 << "]\n";
#endif

	m_InvertCH0_1 = TRUE;
	Params.QueryValue(L"InvertCH0_1", &m_InvertCH0_1);

#if DBG
	t << "m_InvertCH0_1 loaded from registry, resulting value: [" << m_InvertCH0_1 << "]\n";
#endif

	m_InvertCH0_2 = TRUE;
	Params.QueryValue(L"InvertCH0_2", &m_InvertCH0_2);

#if DBG
	t << "m_InvertCH0_2 loaded from registry, resulting value: [" << m_InvertCH0_2 << "]\n";
#endif

	m_InvertCH0_3 = TRUE;
	Params.QueryValue(L"InvertCH0_3", &m_InvertCH0_3);

#if DBG
	t << "m_InvertCH0_3 loaded from registry, resulting value: [" << m_InvertCH0_3 << "]\n";
#endif

	m_InvertCH1_0 = TRUE;
	Params.QueryValue(L"InvertCH1_0", &m_InvertCH1_0);

#if DBG
	t << "m_InvertCH1_0 loaded from registry, resulting value: [" << m_InvertCH1_0 << "]\n";
#endif

	m_InvertCH1_1 = TRUE;
	Params.QueryValue(L"InvertCH1_1", &m_InvertCH1_1);

#if DBG
	t << "m_InvertCH1_1 loaded from registry, resulting value: [" << m_InvertCH1_1 << "]\n";
#endif

	m_InvertCH1_2 = TRUE;
	Params.QueryValue(L"InvertCH1_2", &m_InvertCH1_2);

#if DBG
	t << "m_InvertCH1_2 loaded from registry, resulting value: [" << m_InvertCH1_2 << "]\n";
#endif

	m_InvertCH1_3 = TRUE;
	Params.QueryValue(L"InvertCH1_3", &m_InvertCH1_3);

#if DBG
	t << "m_InvertCH1_3 loaded from registry, resulting value: [" << m_InvertCH1_3 << "]\n";
#endif

	m_CH01SamplingDivisor	=	1;
	Params.QueryValue(L"Ch01SamplingDivisor", &m_CH01SamplingDivisor);

#if DBG
	t << "m_CH01SamplingDivisor loaded from registry, resulting value: [" << m_CH01SamplingDivisor << "]\n";
#endif
	
	m_CH23SamplingDivisor	=	1000;

	Params.QueryValue(L"Ch23SamplingDivisor", &m_CH23SamplingDivisor);

#if DBG
	t << "m_CH23SamplingDivisor loaded from registry, resulting value: [" << m_CH23SamplingDivisor << "]\n";
#endif

	m_SyncTraceTimerSync = TRUE;
	Params.QueryValue(L"SyncTraceTimerSync", &m_SyncTraceTimerSync);

#if DBG
	t << "m_SyncTraceTimerSync loaded from registry, resulting value: [" << m_SyncTraceTimerSync << "]\n";
#endif

}

////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::Invalidate
//
//	Routine Description:
//		Calls Invalidate methods for system resources
//
//	Parameters:
//		None
//
//	Return Value:
//		None
//
//	Comments:
//		This function is called from InitResources (in error conditions).
//		It calls the Invalidate member funcitons for each resource to free
//		the underlying system resource if allocated.  It is safe to call
//		Invalidate more than once for a resource, or for an uninitialized resource.

VOID NTPCFOMDevice::Invalidate()
{
	// For each I/O port mapped region, release the underlying system resource.
	m_PccIo.Invalidate();
}


////////////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::Add_PccIo
//
//	Routine Description:
//		Add an I/O port region to the resource request
//		under construction.
//
//	Parameters:
//		theRequest - pointer to KResourceRequest under construction
//		Params - Registry key for this device
//
//	Return Value:
//		NTSTATUS - Success or failure status
//
//	Comments:
//

NTSTATUS NTPCFOMDevice::Add_PccIo(KResourceRequest *theRequest, KRegistryKey *Params)
{
    PHYSICAL_ADDRESS theAddr;
	theAddr.HighPart = 0;

	theAddr.LowPart =	m_IoBaseAddress;

// TODO:	Check value of theAddr to assure valid data from registry
//		    <high part><low part> form a 64-bit address,
//			high part is usually 0.

	return theRequest->AddPort(
		theAddr,
		theAddr,
		8,
		1,
		0,
		CmResourceShareDeviceExclusive
		);
}

////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::InitResources
//
//	Routine Description:
//		Claim and initialize hardware resources for this device
//
//	Parameters:
//		None
//
//	Return Value:
//		NTSTATUS - STATUS_SUCCESS, or an error code
//
//	Comments:
//		

NTSTATUS NTPCFOMDevice::InitResources(void)
{
	NTSTATUS status;

	KRegistryKey Params(*m_RegPath);
	
	if (!NT_SUCCESS(Params.LastError()))
	{
		// Error, cannot open device registry key
#if DBG
		t << "Failed to create registry key: " << PWSTR(*m_RegPath) << EOL;
		t << "(Did you run REGINI on the .ini file created by the wizard?)\n";
#endif	
		return STATUS_UNSUCCESSFUL;			
	}

	// Create a resource request object, then call 'Add' routines to
	// add requests for each of the resources required.
#if DBG
	t << "Building resource request for ISA device\n";
#endif	
	KResourceRequest theRequest(Isa, (ULONG) 0, (ULONG) 0);

	Add_PccIo(&theRequest, &Params);

    // Submit the request to the system
    status = theRequest.Submit(this, *m_RegPath);

	if (!NT_SUCCESS(status)) 
		return status;

	// For each port I/O mapped region, setup a KResourceAssignment to probe
	// the assigned resource. Initialize the port I/O mapped range using
	// the start address and length obtained from the instance of
	// KResourceAssignment. Once initialized, use member functions such as
	// inb/outb, or the array element operator to access the port range.
	KResourceAssignment AssignedPccIo(&theRequest, CmResourceTypePort, 0);

	// Confirm that the assignment was successful
	if (!AssignedPccIo.Status())
	{
		// Error, could not assign required resource
		Invalidate();
		return STATUS_UNSUCCESSFUL;
	}

	status = m_PccIo.Initialize(
		Isa,
		0,
		AssignedPccIo.Start().QuadPart,
		AssignedPccIo.Length(),
		TRUE
		);

	if (!NT_SUCCESS(status))
	{
		Invalidate();
		return status;		
	}

	m_PortRegData		=	m_PccIo[PC_PCCIO_REG_DATA];
	m_PortRegStatus		=	m_PccIo[PC_PCCIO_REG_STATUS];
	m_PortRegIrqRes		=	m_PccIo[PC_PCCIO_REG_IRQ_RES];
	m_PortRegMode1		=	m_PccIo[PC_PCCIO_REG_MODE1];
	m_PortRegMode2		=	m_PccIo[PC_PCCIO_REG_MODE2];
	m_PortRegReset		=	m_PccIo[PC_PCCIO_REG_RESET];
	m_PortRegDoIrq		=	m_PccIo[PC_PCCIO_REG_DO_IRQ];

	status	=	LoadDeviceFirmware();

	if (!NT_SUCCESS(status))
	{
		Invalidate();
		return status;		
	}

	return STATUS_SUCCESS;
}

#pragma code_seg()

////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::SerialDeviceControl
//	
//	Routine Description:
//		Handler for serialized DeviceControl
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		None
//
//	Comments:
//		This routine is called when the IRP is removed from the
//		queue.  This guarantees that multiple requests are
//		never processed simultaneously.

void NTPCFOMDevice::SerialDeviceControl(KIrp I)
{
#if DBG
	t << "Entering NTPCFOMDevice::SerialWrite, " << I;
#endif

	NTSTATUS status		= STATUS_SUCCESS;

	PUCHAR	pBuffer		= (PUCHAR)I.BufferedWriteSource();

	ULONG   dwTotalSize = I.WriteSize(CURRENT);
	ULONG   dwBytesSent = 0;

// TODO:	If the write can be satisfied immediately, set the Information
//			and Status fields now, then call NextIrp to complete this IRP
//			and start processing the next IRP in the queue.

// TODO:	If the device cannot accept all of the data yet, initiate a 
//			request to the physical device here, and defer the Information,
//			Status, and NextIrp handling until the hardware indicates that
//			the write is complete.  Typically, this might be handled in a
//			DPC that is called after the hardware finishes transferring
//			the data.

// TODO:	To satisfy the write now, transfer data to the device
//			from caller's buffer at "pBuffer".  Then, indicate how much
//			data was transferred:

	I.Information() = dwBytesSent;

	I.Status() = status;


	// NextIrp completes this IRP and starts processing 
	// for the next IRP in the driver managed queue.
// TODO:	The Wizard creates a single queue for all Irps.
//			If you have created additional queues, select
//			the appropriate queue for this Irp here.
	m_DriverManagedQueue.NextIrp(I);
}


////////////////////////////////////////////////////////////////////
//	StartIo
// 
VOID NTPCFOMDevice::StartIo(KIrp I)
{
#if DBG
	t << "Entering StartIo: " << I << "\n";
#endif

	// Before processing the Irp, we need to check to see if it has been
	// cancelled and set it to a non-cancellable state.  When performing
	// these operations, it is necessary to hold the global cancel spin
	// lock and take special precautions to ensure the Irp is still valid.
	// This is accomplished using the routine KIrp::TestAndSetCancelRoutine().
	if ( !I.TestAndSetCancelRoutine(
		LinkTo(CancelQueuedIrp),
		NULL,
		CurrentIrp()) )
	{
		
		// The Irp has been cancelled we stop processing and exit.  Since
		// it was in a cancellable state previously, it will be completed by
		// the cancel routine that had been set on it.
		return;
	}

	// Start processing request. Switch on the IRP's function:
	switch (I.MajorFunction())
	{
		case IRP_MJ_DEVICE_CONTROL:

			break;

		default:
			// Error - unexpected IRP received
			break;
	}
}

////////////////////////////////////////////////////////////////////
//	CancelQueuedIrp
//		This routine is called when a queued IRP is cancelled. The
//		global cancel spin lock is held, and must be released upon
//		exit.
//
VOID NTPCFOMDevice::CancelQueuedIrp(KIrp I)
{
#if DBG
	t << "Entering CancelQueuedIrp: " << I << "\n";
#endif

	KDeviceQueue dq(DeviceQueue());

	if ( (PIRP)I == CurrentIrp() )
	{
		CurrentIrp() = NULL;
		CancelSpinLock::Release(I.CancelIrql());
		I.Information() = 0;
		I.Status() = STATUS_CANCELLED;
		NextIrp(I, IO_NO_INCREMENT, TRUE);
	}
	else if (dq.RemoveSpecificEntry(I))
	{
		CancelSpinLock::Release(I.CancelIrql());
		I.Information() = 0;
		I.Complete(STATUS_CANCELLED);
	}
	else
	{
		CancelSpinLock::Release(I.CancelIrql());
	}
}

///////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::Cleanup
//
//	Routine Description:
//		Handler for IRP_MJ_CLEANUP	
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS	Result code
//
//	Comments:
//

NTSTATUS NTPCFOMDevice::CleanUp(KIrp I)
{
#if DBG
	t << "Entering CleanUp, " << I;
#endif

// TODO:	Insert your code to respond to the CLEANUP message.
//			This code cleans up the single Wizard created queue.  If you
//			have created additional queues,	or have any outstanding Irps
//			stored in some other fashion in your driver, you should clean
//			these up as well for the file object specified in the cleanup Irp.

	m_DriverManagedQueue.CleanUp(I.FileObject());
	return I.Complete(STATUS_SUCCESS);
}


////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::Create
//
//	Routine Description:
//		Handler for IRP_MJ_CREATE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//

NTSTATUS NTPCFOMDevice::Create(KIrp I)
{
#if DBG
	t << "Entering Create, " << I;
#endif

	I.Information() = 0;
	return I.Complete(STATUS_SUCCESS);
}

////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::Close
//
//	Routine Description:
//		Handler for IRP_MJ_CLOSE
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS - Result code
//
//	Comments:
//

NTSTATUS NTPCFOMDevice::Close(KIrp I)
{
#if DBG
	t << "Entering Close, " << I;
#endif
	I.Information() = 0;
	return I.Complete(STATUS_SUCCESS);
}

////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice::Shutdown
//
//	Routine Description:
//		Handler for IRP_MJ_SHUTDOWN
//
//	Parameters:
//		I - Current IRP
//
//	Return Value:
//		NTSTATUS	Result code	
//
//	Comments:
//

NTSTATUS NTPCFOMDevice::Shutdown(KIrp I)
{
#if DBG
	t << "Entering Shutdown\n";
#endif
	// TODO:	Insert SHUTDOWN code

   return I.Complete(STATUS_SUCCESS);
}


////////////////////////////////////////////////////////////////////////////////
//  NTPCFOMDevice_DriverManagedQueue::StartIo
//	
//	Routine Description:
//		This routine is called when an IRP is taken off
//		the Driver Managed Queue (used for serializing I/O) and
//		presented for processing.
//
//	Parameters:
//		I - IRP removed from queue
//
//	Return Value:
//		None
//
//	Comments:
//

VOID NTPCFOMDevice_DriverManagedQueue::StartIo(KIrp I)
{
#if DBG
	t  << "Entering NTPCFOMDevice_DriverManagedQueue StartIo, " << I;
#endif
	// The KDriverManagedQueueEx class gives us the Irp in a non-cancelable state
	// (cancel routine set to NULL) so we can process it without having to worry
	// about clearing the cancel routine first, as is the case with system queuing,
	// or the legacy class KDriverManagedQueue. You may want to set a different cancel
	// routine here, or at other points during the processing of this Irp.

	// Find the device class so we can call the serialized
	// routines in the device class.  The handlers can be
	// moved to the DriverManagedQueue class if it is more
	// convenient.
	NTPCFOMDevice *pDev = (NTPCFOMDevice *) KDevicePTR(I.DeviceObject());

	// Start processing request.

	// Switch on the IRP's function:
	switch (I.MajorFunction())
	{
		case IRP_MJ_DEVICE_CONTROL:
			switch (I.IoctlCode())
			{
				default:
					// We queued a request that shouldn't have been queued
					// (should never get here)
					ASSERT(FALSE);
					break;
			}
			break;

		default:
			// Error - unexpected IRP received
			// NextIrp completes this IRP and starts processing
			// for the next IRP in the queue.
			ASSERT(FALSE);
			I.Status() = STATUS_INVALID_PARAMETER;
			NextIrp(I);
			break;
	}
}

